@@ -1,5 +1,8 @@
 /* fhandler.cc.  See console.cc for fhandler_console functions.
 
+   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006,
+   2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Red Hat, Inc.
+
 This file is part of Cygwin.
 
 This software is a copyrighted work licensed under the terms of the
@@ -10,7 +13,7 @@ details. */
 #include <unistd.h>
 #include <stdlib.h>
 #include <sys/uio.h>
-#include <cygwin/acl.h>
+#include <sys/acl.h>
 #include <sys/param.h>
 #include "cygerrno.h"
 #include "perprocess.h"
@@ -30,10 +33,28 @@ details. */
 #include <asm/socket.h>
 #include "cygwait.h"
 
+#define MAX_OVERLAPPED_WRITE_LEN (64 * 1024 * 1024)
+#define MIN_OVERLAPPED_WRITE_LEN (1 * 1024 * 1024)
+
 static const int CHUNK_SIZE = 1024; /* Used for crlf conversions */
 
 struct __cygwin_perfile *perfile_table;
 
+HANDLE NO_COPY fhandler_base_overlapped::asio_done;
+LONG NO_COPY fhandler_base_overlapped::asio_close_counter;
+
+void
+fhandler_base::reset (const fhandler_base *from)
+{
+  pc << from->pc;
+  rabuf = NULL;
+  ralen = 0;
+  raixget = 0;
+  raixput = 0;
+  rabuflen = 0;
+  _refcnt = 0;
+}
+
 int
 fhandler_base::puts_readahead (const char *s, size_t len)
 {
@@ -48,15 +69,15 @@ int
 fhandler_base::put_readahead (char value)
 {
   char *newrabuf;
-  if (raixput () < rabuflen ())
+  if (raixput < rabuflen)
     /* Nothing to do */;
-  else if ((newrabuf = (char *) realloc (rabuf (), rabuflen () += 32)))
-    rabuf () = newrabuf;
+  else if ((newrabuf = (char *) realloc (rabuf, rabuflen += 32)))
+    rabuf = newrabuf;
   else
     return 0;
 
-  rabuf ()[raixput ()++] = value;
-  ralen ()++;
+  rabuf[raixput++] = value;
+  ralen++;
   return 1;
 }
 
@@ -64,11 +85,11 @@ int
 fhandler_base::get_readahead ()
 {
   int chret = -1;
-  if (raixget () < ralen ())
-    chret = ((unsigned char) rabuf ()[raixget ()++]) & 0xff;
+  if (raixget < ralen)
+    chret = ((unsigned char) rabuf[raixget++]) & 0xff;
   /* FIXME - not thread safe */
-  if (raixget () >= ralen ())
-    raixget () = raixput () = ralen () = 0;
+  if (raixget >= ralen)
+    raixget = raixput = ralen = 0;
   return chret;
 }
 
@@ -76,10 +97,10 @@ int
 fhandler_base::peek_readahead (int queryput)
 {
   int chret = -1;
-  if (!queryput && raixget () < ralen ())
-    chret = ((unsigned char) rabuf ()[raixget ()]) & 0xff;
-  else if (queryput && raixput () > 0)
-    chret = ((unsigned char) rabuf ()[raixput () - 1]) & 0xff;
+  if (!queryput && raixget < ralen)
+    chret = ((unsigned char) rabuf[raixget]) & 0xff;
+  else if (queryput && raixput > 0)
+    chret = ((unsigned char) rabuf[raixput - 1]) & 0xff;
   return chret;
 }
 
@@ -87,11 +108,31 @@ void
 fhandler_base::set_readahead_valid (int val, int ch)
 {
   if (!val)
-    ralen () = raixget () = raixput () = 0;
+    ralen = raixget = raixput = 0;
   if (ch != -1)
     put_readahead (ch);
 }
 
+int
+fhandler_base::eat_readahead (int n)
+{
+  int oralen = ralen;
+  if (n < 0)
+    n = ralen;
+  if (n > 0 && ralen)
+    {
+      if ((int) (ralen -= n) < 0)
+	ralen = 0;
+
+      if (raixget >= ralen)
+	raixget = raixput = ralen = 0;
+      else if (raixput > ralen)
+	raixput = ralen;
+    }
+
+  return oralen;
+}
+
 int
 fhandler_base::get_readahead_into_buffer (char *buf, size_t buflen)
 {
@@ -119,24 +160,10 @@ fhandler_base::set_name (path_conv &in_pc)
 
 char *fhandler_base::get_proc_fd_name (char *buf)
 {
-  IO_STATUS_BLOCK io;
-  FILE_STANDARD_INFORMATION fsi;
-
-  /* If the file had been opened with O_TMPFILE, don't expose the filename. */
-  if ((get_flags () & O_TMPFILE)
-      || (get_device () == FH_FS
-	  && NT_SUCCESS (NtQueryInformationFile (get_handle (), &io,
-						 &fsi, sizeof fsi,
-						 FileStandardInformation))
-	  && fsi.DeletePending))
-    {
-      stpcpy (stpcpy (buf, get_name ()), " (deleted)");
-      return buf;
-    }
   if (get_name ())
     return strcpy (buf, get_name ());
-  if (dev ().name ())
-    return strcpy (buf, dev ().name ());
+  if (dev ().name)
+    return strcpy (buf, dev ().name);
   return strcpy (buf, "");
 }
 
@@ -178,6 +205,8 @@ fhandler_base::set_flags (int flags, int supplied_bin)
     bin = wbinary () || rbinary () ? O_BINARY : O_TEXT;
 
   openflags = flags | bin;
+  if (openflags & O_NONBLOCK)
+    was_nonblocking (true);
 
   bin &= O_BINARY;
   rbinary (bin ? true : false);
@@ -235,7 +264,7 @@ retry:
 		  break;
 		}
 	    }
-	  fallthrough;
+	  /*FALLTHRU*/
 	case STATUS_INVALID_DEVICE_REQUEST:
 	case STATUS_INVALID_PARAMETER:
 	case STATUS_INVALID_HANDLE:
@@ -245,7 +274,6 @@ retry:
 	      len = (size_t) -1;
 	      break;
 	    }
-	  fallthrough;
 	default:
 	  __seterrno_from_nt_status (status);
 	  len = (size_t) -1;
@@ -316,9 +344,6 @@ fhandler_base::device_access_denied (int flags)
 {
   int mode = 0;
 
-  if (flags & O_PATH)
-    return false;
-
   if (flags & O_RDWR)
     mode |= R_OK | W_OK;
   if (flags & (O_WRONLY | O_APPEND))
@@ -445,9 +470,9 @@ fhandler_base::open_with_arch (int flags, mode_t mode)
     }
   else if (archetype)
     {
-      if (!archetype->get_handle ())
+      if (!archetype->get_io_handle ())
 	{
-	  archetype->copy_from (this);
+	  copyto (archetype);
 	  archetype_usecount (1);
 	  archetype->archetype = NULL;
 	  usecount = 0;
@@ -464,15 +489,14 @@ fhandler_base::open_with_arch (int flags, mode_t mode)
 	      strcpy (name, get_name ());
 	    }
 	  fhandler_base *arch = archetype;
-	  copy_from (archetype);
+	  archetype->copyto (this);
 	  if (name)
 	    set_name (name);
 	  archetype = arch;
 	  archetype_usecount (1);
 	  usecount = 0;
 	}
-      if (!open_setup (flags))
-	api_fatal ("open_setup failed, %E");
+      open_setup (flags);
     }
 
   close_on_exec (flags & O_CLOEXEC);
@@ -507,7 +531,7 @@ fhandler_base::open_null (int flags)
       __seterrno_from_nt_status (status);
       goto done;
     }
-  set_handle (fh);
+  set_io_handle (fh);
   set_flags (flags, pc.binmode ());
   res = 1;
   set_open_status ();
@@ -532,24 +556,9 @@ fhandler_base::open (int flags, mode_t mode)
   PFILE_FULL_EA_INFORMATION p = NULL;
   ULONG plen = 0;
 
-  syscall_printf ("(%S, %y)%s", pc.get_nt_native_path (), flags,
-				get_handle () ? " by handle" : "");
+  syscall_printf ("(%S, %y)", pc.get_nt_native_path (), flags);
 
-  if (flags & O_PATH)
-    query_open (query_read_attributes);
-
-  /* Allow to reopen from handle.  This is utilized by
-     open ("/proc/PID/fd/DESCRIPTOR", ...); */
-  if (get_handle ())
-    {
-      pc.init_reopen_attr (attr, get_handle ());
-      if (!(flags & O_CLOEXEC))
-	attr.Attributes |= OBJ_INHERIT;
-      if (pc.has_buggy_reopen ())
-	debug_printf ("Reopen by handle requested but FS doesn't support it");
-    }
-  else
-    pc.get_object_attr (attr, *sec_none_cloexec (flags));
+  pc.get_object_attr (attr, *sec_none_cloexec (flags));
 
   options = FILE_OPEN_FOR_BACKUP_INTENT;
   switch (query_open ())
@@ -561,12 +570,10 @@ fhandler_base::open (int flags, mode_t mode)
       access = READ_CONTROL | FILE_READ_ATTRIBUTES;
       break;
     case query_write_control:
-      access = READ_CONTROL | WRITE_OWNER | WRITE_DAC
-	       | (pc.fs_is_samba () ? 0 : FILE_WRITE_ATTRIBUTES);
+      access = READ_CONTROL | WRITE_OWNER | WRITE_DAC | FILE_WRITE_ATTRIBUTES;
       break;
     case query_write_dac:
-      access = READ_CONTROL | WRITE_DAC
-	       | (pc.fs_is_samba () ? 0 : FILE_WRITE_ATTRIBUTES);
+      access = READ_CONTROL | WRITE_DAC | FILE_WRITE_ATTRIBUTES;
       break;
     case query_write_attributes:
       access = READ_CONTROL | FILE_WRITE_ATTRIBUTES;
@@ -598,39 +605,17 @@ fhandler_base::open (int flags, mode_t mode)
 
   /* Don't use the FILE_OVERWRITE{_IF} flags here.  See below for an
      explanation, why that's not such a good idea. */
-  if (((flags & O_EXCL) && (flags & O_CREAT)) || (flags & O_TMPFILE))
+  if ((flags & O_EXCL) && (flags & O_CREAT))
     create_disposition = FILE_CREATE;
   else
     create_disposition = (flags & O_CREAT) ? FILE_OPEN_IF : FILE_OPEN;
 
-  if (get_device () == FH_FS
-#ifdef __WITH_AF_UNIX
-      || get_device () == FH_UNIX
-#endif
-      )
-    {
-      /* Add the reparse point flag to known reparse points, otherwise we
-	 open the target, not the reparse point.  This would break lstat. */
-      if (pc.is_known_reparse_point ())
-	options |= FILE_OPEN_REPARSE_POINT;
-    }
-
   if (get_device () == FH_FS)
     {
-      /* O_TMPFILE files are created with delete-on-close semantics, as well
-	 as with FILE_ATTRIBUTE_TEMPORARY.  The latter speeds up file access,
-	 because the OS tries to keep the file in memory as much as possible.
-	 In conjunction with FILE_DELETE_ON_CLOSE, ideally the OS never has
-	 to write to the disk at all.
-	 Note that O_TMPFILE_FILE_ATTRS also sets the DOS HIDDEN attribute
-	 to help telling Cygwin O_TMPFILE files apart from other files
-	 accidentally setting FILE_ATTRIBUTE_TEMPORARY. */
-      if (flags & O_TMPFILE)
-	{
-	  access |= DELETE;
-	  file_attributes |= O_TMPFILE_FILE_ATTRS;
-	  options |= FILE_DELETE_ON_CLOSE;
-	}
+      /* Add the reparse point flag to native symlinks, otherwise we open the
+	 target, not the symlink.  This would break lstat. */
+      if (pc.is_rep_symlink ())
+	options |= FILE_OPEN_REPARSE_POINT;
 
       if (pc.fs_is_nfs ())
 	{
@@ -647,7 +632,15 @@ fhandler_base::open (int flags, mode_t mode)
 	    }
 	}
 
-      if (flags & (O_CREAT | O_TMPFILE))
+      /* Trying to overwrite an already existing file with FILE_ATTRIBUTE_HIDDEN
+	 and/or FILE_ATTRIBUTE_SYSTEM attribute set, NtCreateFile fails with
+	 STATUS_ACCESS_DENIED.  Per MSDN you have to create the file with the
+	 same attributes as already specified for the file. */
+      if (create_disposition == FILE_CREATE
+	  && has_attribute (FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM))
+	file_attributes |= pc.file_attributes ();
+
+      if (flags & O_CREAT)
 	{
 	  file_attributes |= FILE_ATTRIBUTE_NORMAL;
 
@@ -690,23 +683,6 @@ fhandler_base::open (int flags, mode_t mode)
 
   status = NtCreateFile (&fh, access, &attr, &io, NULL, file_attributes, shared,
 			 create_disposition, options, p, plen);
-  /* Pre-W10, we can't reopen a file by handle with delete disposition
-     set, so we have to lie our ass off. */
-  if (get_handle () && status == STATUS_DELETE_PENDING)
-    {
-      BOOL ret = DuplicateHandle (GetCurrentProcess (), get_handle (),
-				  GetCurrentProcess (), &fh,
-				  access, !(flags & O_CLOEXEC), 0);
-      if (!ret)
-	ret = DuplicateHandle (GetCurrentProcess (), get_handle (),
-			       GetCurrentProcess (), &fh,
-			       0, !(flags & O_CLOEXEC),
-			       DUPLICATE_SAME_ACCESS);
-      if (!ret)
-	debug_printf ("DuplicateHandle after STATUS_DELETE_PENDING, %E");
-      else
-	status = STATUS_SUCCESS;
-    }
   if (!NT_SUCCESS (status))
     {
       /* Trying to create a directory should return EISDIR, not ENOENT. */
@@ -769,7 +745,7 @@ fhandler_base::open (int flags, mode_t mode)
 	}
     }
 
-  set_handle (fh);
+  set_io_handle (fh);
   set_flags (flags, pc.binmode ());
 
   res = 1;
@@ -785,17 +761,9 @@ done:
   return res;
 }
 
-fhandler_base *
-fhandler_base::fd_reopen (int, mode_t)
-{
-  /* This is implemented in fhandler_process only. */
-  return NULL;
-}
-
-bool
+void
 fhandler_base::open_setup (int)
 {
-  return true;
 }
 
 /* states:
@@ -868,7 +836,7 @@ out:
 ssize_t __stdcall
 fhandler_base::write (const void *ptr, size_t len)
 {
-  ssize_t res;
+  int res;
 
   if (did_lseek ())
     {
@@ -941,9 +909,9 @@ fhandler_base::write (const void *ptr, size_t len)
 	    }
 
 	  /* We've got a buffer-full, or we're out of data.  Write it out */
-	  ssize_t nbytes;
-	  ptrdiff_t want = buf_ptr - buf;
-	  if ((nbytes = raw_write (buf, (size_t) want)) == want)
+	  int nbytes;
+	  int want = buf_ptr - buf;
+	  if ((nbytes = raw_write (buf, want)) == want)
 	    {
 	      /* Keep track of how much written not counting additional \r's */
 	      res = data - (char *)ptr;
@@ -1084,7 +1052,7 @@ fhandler_base::lseek (off_t offset, int whence)
   if (whence != SEEK_CUR || offset != 0)
     {
       if (whence == SEEK_CUR)
-	offset -= ralen () - raixget ();
+	offset -= ralen - raixget;
       set_readahead_valid (0);
     }
 
@@ -1134,20 +1102,20 @@ fhandler_base::lseek (off_t offset, int whence)
      readahead that we have to take into account when calculating
      the actual position for the application.  */
   if (whence == SEEK_CUR)
-    res -= ralen () - raixget ();
+    res -= ralen - raixget;
 
   return res;
 }
 
 ssize_t __reg3
-fhandler_base::pread (void *, size_t, off_t, void *)
+fhandler_base::pread (void *, size_t, off_t)
 {
   set_errno (ESPIPE);
   return -1;
 }
 
 ssize_t __reg3
-fhandler_base::pwrite (void *, size_t, off_t, void *)
+fhandler_base::pwrite (void *, size_t, off_t)
 {
   set_errno (ESPIPE);
   return -1;
@@ -1218,6 +1186,88 @@ fhandler_base::close ()
   return res;
 }
 
+DWORD WINAPI
+flush_async_io (void *arg)
+{
+  fhandler_base_overlapped *fh = (fhandler_base_overlapped *) arg;
+  debug_only_printf ("waiting for write I/O for %s", fh->get_name ());
+  DWORD nbytes;
+  bool res = GetOverlappedResult (fh->get_output_handle (),
+				  fh->get_overlapped (), &nbytes, true);
+  debug_printf ("finished waiting for I/O from %s, res %d", fh->get_name (),
+		res);
+  fh->close ();
+  delete fh;
+
+  InterlockedDecrement (&fhandler_base_overlapped::asio_close_counter);
+  SetEvent (fhandler_base_overlapped::asio_done);
+
+  _my_tls._ctinfo->auto_release ();
+  return 0;
+}
+
+void
+fhandler_base_overlapped::flush_all_async_io ()
+{
+  while (asio_close_counter > 0)
+    if (WaitForSingleObject (asio_done, INFINITE) != WAIT_OBJECT_0)
+      {
+	system_printf ("WaitForSingleObject failed, possible data loss in pipe, %E");
+	break;
+      }
+  asio_close_counter = 0;
+  if (asio_done)
+    CloseHandle (asio_done);
+}
+
+/* Start a thread to handle closing of overlapped asynchronous I/O since
+   Windows amazingly does not seem to always flush I/O on close.  */
+void
+fhandler_base_overlapped::check_later ()
+{
+  set_close_on_exec (true);
+  char buf[MAX_PATH];
+  if (!asio_done
+      && !(asio_done = CreateEvent (&sec_none_nih, false, false,
+				    shared_name (buf, "asio",
+						 GetCurrentProcessId ()))))
+    api_fatal ("CreateEvent failed, %E");
+
+  InterlockedIncrement (&asio_close_counter);
+  if (!new cygthread(flush_async_io, this, "flasio"))
+    api_fatal ("couldn't create a thread to track async I/O, %E");
+  debug_printf ("started thread to handle asynchronous closing for %s", get_name ());
+}
+
+int
+fhandler_base_overlapped::close ()
+{
+  int res;
+  int writer = (get_access () & GENERIC_WRITE);
+  /* Need to treat non-blocking I/O specially because Windows appears to
+     be brain-dead.  We're checking here if the descriptor was ever set
+     to nonblocking, rather than checking if it's nonblocking at close time.
+     The reason is that applications may switch back to blocking (for the
+     sake of some other application accessing this descriptor) without
+     performaing any further I/O.  These applications would suffer data
+     loss, which this workaround is trying to fix. */
+  if (writer && was_nonblocking () && has_ongoing_io ())
+    {
+      clone (HEAP_3_FHANDLER)->check_later ();
+      res = 0;
+    }
+  else
+    {
+     /* Cancelling seems to be necessary for cases where a reader is
+	 still executing when a signal handler performs a close.  */
+      if (!writer)
+	CancelIo (get_io_handle ());
+      destroy_overlapped ();
+      res = fhandler_base::close ();
+    }
+  return res;
+}
+
 int
 fhandler_base::ioctl (unsigned int cmd, void *buf)
 {
@@ -1295,7 +1345,7 @@ fhandler_base::fstatvfs (struct statvfs *sfs)
 int
 fhandler_base::init (HANDLE f, DWORD a, mode_t bin)
 {
-  set_handle (f);
+  set_io_handle (f);
   access = a;
   a &= GENERIC_READ | GENERIC_WRITE;
   int flags = 0;
@@ -1312,7 +1362,7 @@ fhandler_base::init (HANDLE f, DWORD a, mode_t bin)
 }
 
 int
-fhandler_base::dup (fhandler_base *child, int flags)
+fhandler_base::dup (fhandler_base *child, int)
 {
   debug_printf ("in fhandler_base dup");
 
@@ -1321,7 +1371,7 @@ fhandler_base::dup (fhandler_base *child, int flags)
     {
       if (!DuplicateHandle (GetCurrentProcess (), get_handle (),
 			    GetCurrentProcess (), &nh,
-			    0, !(flags & O_CLOEXEC), DUPLICATE_SAME_ACCESS))
+			    0, TRUE, DUPLICATE_SAME_ACCESS))
 	{
 	  debug_printf ("dup(%s) failed, handle %p, %E",
 			get_name (), get_handle ());
@@ -1330,11 +1380,19 @@ fhandler_base::dup (fhandler_base *child, int flags)
 	}
 
       VerifyHandle (nh);
-      child->set_handle (nh);
+      child->set_io_handle (nh);
     }
   return 0;
 }
 
+int
+fhandler_base_overlapped::dup (fhandler_base *child, int flags)
+{
+  int res = fhandler_base::dup (child, flags) ||
+	    ((fhandler_base_overlapped *) child)->setup_overlapped ();
+  return res;
+}
+
 int fhandler_base::fcntl (int cmd, intptr_t arg)
 {
   int res;
@@ -1467,24 +1525,23 @@ fhandler_base::fhandler_base () :
   ino (0),
   _refcnt (0),
   openflags (0),
+  rabuf (NULL),
+  ralen (0),
+  raixget (0),
+  raixput (0),
+  rabuflen (0),
   unique_id (0),
-  select_sem (NULL),
   archetype (NULL),
   usecount (0)
 {
-  ra.rabuf = NULL;
-  ra.ralen = 0;
-  ra.raixget = 0;
-  ra.raixput = 0;
-  ra.rabuflen = 0;
   isclosed (false);
 }
 
 /* Normal I/O destructor */
 fhandler_base::~fhandler_base ()
 {
-  if (ra.rabuf)
-    free (ra.rabuf);
+  if (rabuf)
+    free (rabuf);
 }
 
 /**********************************************************************/
@@ -1495,15 +1552,6 @@ fhandler_dev_null::fhandler_dev_null () :
 {
 }
 
-ssize_t __stdcall
-fhandler_dev_null::write (const void *ptr, size_t len)
-{
-  /* Shortcut.  This also fixes a problem with the NUL device on 64 bit:
-     If you write > 4 GB in a single attempt, the bytes written returned
-     from by is numBytes & 0xffffffff. */
-  return len;
-}
-
 void
 fhandler_base::set_no_inheritance (HANDLE &h, bool not_inheriting)
 {
@@ -1521,7 +1569,7 @@ fhandler_base::fork_fixup (HANDLE parent, HANDLE &h, const char *name)
 {
   HANDLE oh = h;
   bool res = false;
-  if (!close_on_exec ())
+  if (/* !is_socket () && */ !close_on_exec ())
     debug_printf ("handle %p already opened", h);
   else if (!DuplicateHandle (parent, h, GetCurrentProcess (), &h,
 			     0, !close_on_exec (), DUPLICATE_SAME_ACCESS))
@@ -1555,6 +1603,13 @@ fhandler_base::fixup_after_fork (HANDLE parent)
     del_my_locks (after_fork);
 }
 
+void
+fhandler_base_overlapped::fixup_after_fork (HANDLE parent)
+{
+  setup_overlapped ();
+  fhandler_base::fixup_after_fork (parent);
+}
+
 void
 fhandler_base::fixup_after_exec ()
 {
@@ -1563,6 +1618,12 @@ fhandler_base::fixup_after_exec ()
     del_my_locks (after_exec);
   mandatory_locking (false);
 }
+void
+fhandler_base_overlapped::fixup_after_exec ()
+{
+  setup_overlapped ();
+  fhandler_base::fixup_after_exec ();
+}
 
 bool
 fhandler_base::is_nonblocking ()
@@ -1576,6 +1637,8 @@ fhandler_base::set_nonblocking (int yes)
   int current = openflags & O_NONBLOCK_MASK;
   int new_flags = yes ? (!current ? O_NONBLOCK : current) : 0;
   openflags = (openflags & ~O_NONBLOCK_MASK) | new_flags;
+  if (new_flags)
+    was_nonblocking (true);
 }
 
 int
@@ -1642,6 +1705,7 @@ fhandler_base::closedir (DIR *)
 int
 fhandler_base::fchmod (mode_t mode)
 {
+  extern int chmod_device (path_conv& pc, mode_t mode);
   if (pc.is_fs_special ())
     return chmod_device (pc, mode);
   /* By default, just succeeds. */
@@ -1721,7 +1785,8 @@ fhandler_base::fadvise (off_t offset, off_t length, int advice)
 int
 fhandler_base::ftruncate (off_t length, bool allow_truncate)
 {
-  return EINVAL;
+  set_errno (EINVAL);
+  return -1;
 }
 
 int
@@ -1813,7 +1878,6 @@ fhandler_base::fpathconf (int v)
       set_errno (EINVAL);
       break;
     case _PC_ASYNC_IO:
-      return 1;
     case _PC_PRIO_IO:
       break;
     case _PC_SYNC_IO:
@@ -1830,8 +1894,6 @@ fhandler_base::fpathconf (int v)
 	return pc.has_acls () || pc.fs_is_nfs ();
       set_errno (EINVAL);
       break;
-    case _PC_CASE_INSENSITIVE:
-      return !!pc.objcaseinsensitive ();
     default:
       set_errno (EINVAL);
       break;
@@ -1839,32 +1901,272 @@ fhandler_base::fpathconf (int v)
   return -1;
 }
 
-NTSTATUS
-fhandler_base::npfs_handle (HANDLE &nph)
+/* Overlapped I/O */
+
+int __reg1
+fhandler_base_overlapped::setup_overlapped ()
 {
-  static NO_COPY SRWLOCK npfs_lock;
-  static NO_COPY HANDLE npfs_dirh;
+  OVERLAPPED *ov = get_overlapped_buffer ();
+  memset (ov, 0, sizeof (*ov));
+  set_overlapped (ov);
+  ov->hEvent = CreateEvent (&sec_none_nih, true, true, NULL);
+  io_pending = false;
+  return ov->hEvent ? 0 : -1;
+}
 
-  NTSTATUS status = STATUS_SUCCESS;
-  OBJECT_ATTRIBUTES attr;
-  IO_STATUS_BLOCK io;
+void __reg1
+fhandler_base_overlapped::destroy_overlapped ()
+{
+  OVERLAPPED *ov = get_overlapped ();
+  if (ov && ov->hEvent)
+    {
+      SetEvent (ov->hEvent);
+      CloseHandle (ov->hEvent);
+      ov->hEvent = NULL;
+    }
+  io_pending = false;
+  get_overlapped () = NULL;
+}
+
+bool __reg1
+fhandler_base_overlapped::has_ongoing_io ()
+{
+  if (!io_pending)
+    return false;
+
+  if (!IsEventSignalled (get_overlapped ()->hEvent))
+    return true;
+  io_pending = false;
+  DWORD nbytes;
+  GetOverlappedResult (get_output_handle (), get_overlapped (), &nbytes, false);
+  return false;
+}
+
+fhandler_base_overlapped::wait_return __reg3
+fhandler_base_overlapped::wait_overlapped (bool inres, bool writing, DWORD *bytes, bool nonblocking, DWORD len)
+{
+  if (!get_overlapped ())
+    return inres ? overlapped_success : overlapped_error;
+
+  wait_return res = overlapped_unknown;
+  DWORD err;
+  if (inres)
+    /* handle below */;
+  else if ((err = GetLastError ()) != ERROR_IO_PENDING)
+    res = overlapped_error;
+  else if (!nonblocking)
+    /* handle below */;
+  else if (!writing)
+    SetEvent (get_overlapped ()->hEvent);	/* Force immediate WFMO return */
+  else
+    {
+      *bytes = len;				/* Assume that this worked */
+      io_pending = true;			/*  but don't allow subsequent */
+      res = overlapped_success;			/*  writes until completed */
+    }
+  if (res == overlapped_unknown)
+    {
+      DWORD wfres = cygwait (get_overlapped ()->hEvent);
+      HANDLE h = writing ? get_output_handle () : get_handle ();
+      BOOL wores;
+      if (isclosed ())
+	{
+	  switch (wfres)
+	    {
+	    case WAIT_OBJECT_0:
+	      err = ERROR_INVALID_HANDLE;
+	      break;
+	    case WAIT_SIGNALED:
+	      err = ERROR_INVALID_AT_INTERRUPT_TIME;
+	      break;
+	    default:
+	      err = GetLastError ();
+	      break;
+	    }
+	  res = overlapped_error;
+	}
+      else
+	{
+	  /* Cancelling here to prevent races.  It's possible that the I/O has
+	     completed already, in which case this is a no-op.  Otherwise,
+	     WFMO returned because 1) This is a non-blocking call, 2) a signal
+	     arrived, or 3) the operation was cancelled.  These cases may be
+	     overridden by the return of GetOverlappedResult which could detect
+	     that I/O completion occurred.  */
+	  CancelIo (h);
+	  wores = GetOverlappedResult (h, get_overlapped (), bytes, false);
+	  err = GetLastError ();
+	  ResetEvent (get_overlapped ()->hEvent);	/* Probably not needed but CYA */
+	  debug_printf ("wfres %u, wores %d, bytes %u", wfres, wores, *bytes);
+	  if (wores)
+	    res = overlapped_success;	/* operation succeeded */
+	  else if (wfres == WAIT_OBJECT_0 + 1)
+	    {
+	      err = ERROR_INVALID_AT_INTERRUPT_TIME; /* forces an EINTR below */
+	      debug_printf ("signal");
+	      res = overlapped_error;
+	    }
+	  else if (wfres == WAIT_OBJECT_0 + 2)
+	    pthread::static_cancel_self ();		/* never returns */
+	  else if (nonblocking)
+	    res = overlapped_nonblocking_no_data;	/* more handling below */
+	  else
+	    {
+	      debug_printf ("GetOverlappedResult failed, h %p, bytes %u, %E", h, *bytes);
+	      res = overlapped_error;
+	    }
+	}
+    }
+
+  if (res == overlapped_success)
+    {
+      debug_printf ("normal %s, %u bytes ispipe() %d", writing ? "write" : "read", *bytes, ispipe ());
+      if (*bytes == 0 && !writing && ispipe ())
+	res = overlapped_nullread;
+    }
+  else if (res == overlapped_nonblocking_no_data)
+    {
+      *bytes = (DWORD) -1;
+      set_errno (EAGAIN);
+      debug_printf ("no data to read for nonblocking I/O");
+    }
+  else if (err == ERROR_HANDLE_EOF || err == ERROR_BROKEN_PIPE)
+    {
+      debug_printf ("EOF, %E");
+      *bytes = 0;
+      res = overlapped_success;
+      if (writing && err == ERROR_BROKEN_PIPE)
+	raise (SIGPIPE);
+    }
+  else
+    {
+      debug_printf ("res %u, Win32 Error %u", (unsigned) res, err);
+      *bytes = (DWORD) -1;
+      __seterrno_from_win_error (err);
+      if (writing && err == ERROR_NO_DATA)
+	raise (SIGPIPE);
+    }
+
+  return res;
+}
+
+void __reg3
+fhandler_base_overlapped::raw_read (void *ptr, size_t& len)
+{
+  DWORD nbytes;
+  bool keep_looping;
+  do
+    {
+      bool res = ReadFile (get_handle (), ptr, len, &nbytes,
+			   get_overlapped ());
+      switch (wait_overlapped (res, false, &nbytes, is_nonblocking ()))
+	{
+	case overlapped_nullread:
+	  keep_looping = true;
+	  break;
+	default:	/* Added to quiet gcc */
+	case overlapped_success:
+	case overlapped_error:
+	  keep_looping = false;
+	  break;
+	}
+    }
+  while (keep_looping);
+  len = (nbytes == (DWORD) -1) ? (size_t) -1 : (size_t) nbytes;
+}
 
-  /* Lockless after first call. */
-  if (npfs_dirh)
+ssize_t __reg3
+fhandler_base_overlapped::raw_write (const void *ptr, size_t len)
+{
+  size_t nbytes;
+  if (has_ongoing_io ())
     {
-      nph = npfs_dirh;
-      return STATUS_SUCCESS;
+      set_errno (EAGAIN);
+      nbytes = (size_t) -1;
     }
-  AcquireSRWLockExclusive (&npfs_lock);
-  if (!npfs_dirh)
+  else
     {
-      InitializeObjectAttributes (&attr, &ro_u_npfs, 0, NULL, NULL);
-      status = NtOpenFile (&npfs_dirh, FILE_READ_ATTRIBUTES | SYNCHRONIZE,
-			   &attr, &io, FILE_SHARE_READ | FILE_SHARE_WRITE,
-			   0);
+      size_t chunk;
+      if (!max_atomic_write || len < max_atomic_write)
+	chunk = MIN (len, INT_MAX);
+      else if (is_nonblocking ())
+	chunk = len = max_atomic_write;
+      else
+	chunk = max_atomic_write;
+
+      /* MSDN "WriteFile" contains the following note: "Accessing the output
+         buffer while a write operation is using the buffer may lead to
+	 corruption of the data written from that buffer.  [...]  This can
+	 be particularly problematic when using an asynchronous file handle.
+	 (https://msdn.microsoft.com/en-us/library/windows/desktop/aa365747)
+
+	 MSDN "Synchronous and Asynchronous I/O" contains the following note:
+	 "Do not deallocate or modify [...] the data buffer until all
+	 asynchronous I/O operations to the file object have been completed."
+	 (https://msdn.microsoft.com/en-us/library/windows/desktop/aa365683)
+
+	 This problem is a non-issue for blocking I/O, but it can lead to
+	 problems when using nonblocking I/O.  Consider:
+	 - The application uses a static buffer in repeated calls to
+	   non-blocking write.
+	 - The previous write returned with success, but the overlapped I/O
+	   operation is ongoing.
+	 - The application copies the next set of data to the static buffer,
+	   thus overwriting data still accessed by the previous write call.
+	 --> potential data corruption.
+
+	 What we do here is to allocate a per-fhandler buffer big enough
+	 to perform the maximum atomic operation from, copy the user space
+	 data over to this buffer and then call NtWriteFile on this buffer.
+	 This decouples the write operation from the user buffer and the
+	 user buffer can be reused without data corruption issues.
+
+	 Since no further write can occur while we're still having ongoing
+	 I/O, this should be reasanably safe.
+
+	 Note: We only have proof that this problem actually occurs on Wine
+	 yet.  However, the MSDN language indicates that this may be a real
+	 problem on real Windows as well. */
+      if (is_nonblocking ())
+	{
+	  if (!atomic_write_buf)
+	    atomic_write_buf = cmalloc_abort (HEAP_BUF, max_atomic_write);
+	  ptr = memcpy (atomic_write_buf, ptr, chunk);
+	}
+
+      nbytes = 0;
+      DWORD nbytes_now = 0;
+      /* Write to fd in smaller chunks, accumulating a total.
+	 If there's an error, just return the accumulated total
+	 unless the first write fails, in which case return value
+	 from wait_overlapped(). */
+      while (nbytes < len)
+	{
+	  size_t left = len - nbytes;
+	  size_t len1;
+	  if (left > chunk)
+	    len1 = chunk;
+	  else
+	    len1 = left;
+	  bool res = WriteFile (get_output_handle (), ptr, len1, &nbytes_now,
+				get_overlapped ());
+	  switch (wait_overlapped (res, true, &nbytes_now,
+				   is_nonblocking (), len1))
+	    {
+	    case overlapped_success:
+	      ptr = ((char *) ptr) + chunk;
+	      nbytes += nbytes_now;
+	      /* fall through intentionally */
+	    case overlapped_error:
+	      len = 0;		/* terminate loop */
+	    case overlapped_unknown:
+	    case overlapped_nullread:
+	    case overlapped_nonblocking_no_data:
+	      break;
+	    }
+	}
+      if (!nbytes)
+	nbytes = (nbytes_now == (DWORD) -1) ? (size_t) -1 : nbytes_now;
     }
-  ReleaseSRWLockExclusive (&npfs_lock);
-  if (NT_SUCCESS (status))
-    nph = npfs_dirh;
-  return status;
+  return nbytes;
 }
